from plone import api
from zExceptions import BadRequest

import json
import uuid


def get_data_from_request(request, loadjson):
    if loadjson:
        request.stdin.seek(0)
        return json.load(request.stdin)
    data = request.form
    if "payload" in data:
        data["payload"] = json.loads(data["payload"])
    return data


def get_owner(data=None, request=None, loadjson=None):
    if data is None:
        data = get_data_from_request(request, loadjson)
    owner = data.get("owner", None)
    if owner is None:
        owner = api.user.get_current().getId()
        if owner is None:
            raise BadRequest(f"Property 'owner' is mandatory for anonymous users.")
    return owner


def get_bookmark_from_request(request, loadjson=False):
    data = get_data_from_request(request, loadjson)
    uid = data.get("uid", None)
    if not uid:
        raise BadRequest("Property 'uid' is required")
    try:
        uid = uuid.UUID(uid)
    except ValueError as exc:
        raise BadRequest(f"Property 'uid' is malformed: {exc}")
    group = data.get("group", "")
    owner = get_owner(data=data)
    return owner, uid, group, data.get("queryparams", ""), data.get("payload", {})


def bookmark_dict_to_json_dict(bookmark):
    """Prepare result item for response.

    Enrich more if needed.
    """
    bookmark["uid"] = bookmark["uid"].hex

    bookmarkBrain = api.content.find(UID=bookmark["uid"])
    bookmarkUrl = (
        len(bookmarkBrain) > 0 and bookmarkBrain[0].getObject().absolute_url() or "#"
    )
    bookmark["@id"] = bookmarkUrl
    bookmark["title"] = (
        len(bookmarkBrain) > 0 and bookmarkBrain[0].getObject().title or ""
    )
    bookmark["@type"] = (
        len(bookmarkBrain) > 0 and bookmarkBrain[0].getObject().portal_type or ""
    )
    return bookmark



"""Setup tests for this package."""
from collective.bookmarks.testing import COLLECTIVE_BOOKMARKS_INTEGRATION_TESTING

import unittest


class TestStorage(unittest.TestCase):

    layer = COLLECTIVE_BOOKMARKS_INTEGRATION_TESTING

    def test_storage_soup_installed(self):
        from collective.bookmarks.storage import Bookmarks
        from souper.soup import Soup

        bookmarks = Bookmarks()
        self.assertIsInstance(bookmarks._soup, Soup)

    def test_fetch_one_existing(self):
        from collective.bookmarks.storage import Bookmarks
        from repoze.catalog.query import Eq
        from souper.soup import Record

        import uuid

        bookmarks = Bookmarks()
        data = {"owner": "kim stanley", "uid": uuid.uuid4(), "group": ""}
        record = Record()
        record.attrs["owner"] = data["owner"]
        record.attrs["uid"] = data["uid"]
        record.attrs["group"] = data["group"]
        bookmarks._soup.add(record)

        result = bookmarks._fetch_one(
            Eq("owner", data["owner"])
            & Eq("uid", data["uid"])
            & Eq("group", data["group"])
        )
        self.assertIs(result, record)

    def test_fetch_one_non_existing(self):
        from collective.bookmarks.storage import Bookmarks
        from repoze.catalog.query import Eq

        import uuid

        bookmarks = Bookmarks()
        result = bookmarks._fetch_one(
            Eq("owner", "harry") & Eq("uid", uuid.uuid4()) & Eq("group", "cool stuff")
        )
        self.assertIs(result, None)

    def test_fetch_one_too_many(self):
        from collective.bookmarks.storage import Bookmarks
        from repoze.catalog.query import Eq
        from souper.soup import Record

        import uuid

        bookmarks = Bookmarks()
        data = {"owner": "kim stanley", "uid": uuid.uuid4(), "group": ""}
        record1 = Record()
        record1.attrs["owner"] = data["owner"]
        record1.attrs["uid"] = data["uid"]
        record1.attrs["group"] = data["group"]
        record2 = Record()
        record2.attrs["owner"] = data["owner"]
        record2.attrs["uid"] = data["uid"]
        record2.attrs["group"] = data["group"]
        bookmarks._soup.add(record1)
        bookmarks._soup.add(record2)

        with self.assertRaises(ValueError):
            bookmarks._fetch_one(
                Eq("owner", data["owner"])
                & Eq("uid", data["uid"])
                & Eq("group", data["group"])
            )

    def test_dictify(self):
        from collective.bookmarks.storage import Bookmarks
        from souper.soup import Record

        import math
        import time
        import uuid

        bookmarks = Bookmarks()
        data = {
            "owner": "kim stanley",
            "uid": uuid.uuid4(),
            "group": "",
            "payload": {},
            "created": math.floor(time.time()),
        }
        record = Record()
        record.attrs["owner"] = data["owner"]
        record.attrs["uid"] = data["uid"]
        record.attrs["group"] = data["group"]
        record.attrs["payload"] = data["payload"]
        record.attrs["created"] = data["created"]

        dictified = bookmarks._dictify(record)
        self.assertEqual(data, dictified)

    def test_add_first(self):
        from collective.bookmarks.storage import Bookmarks
        from repoze.catalog.query import Eq

        import math
        import time
        import uuid

        bookmarks = Bookmarks()
        data = {
            "owner": "kim stanley",
            "uid": uuid.uuid4(),
            "group": "",
            "payload": {},
            "created": math.floor(time.time()),
        }
        result = bookmarks.add(
            data["owner"], data["uid"], data["group"], data["payload"]
        )
        self.assertIsNotNone(result)

        record = bookmarks._fetch_one(
            Eq("owner", data["owner"])
            & Eq("uid", data["uid"])
            & Eq("group", data["group"])
        )
        self.assertIn("created", record.attrs)
        record.attrs["created"] = data["created"]
        self.assertEqual(data, bookmarks._dictify(record))

    def test_add_same(self):
        from collective.bookmarks.storage import Bookmarks

        import uuid

        bookmarks = Bookmarks()
        data = {"owner": "kim stanley", "uid": uuid.uuid4(), "group": "", "payload": {}}
        result1 = bookmarks.add(
            data["owner"], data["uid"], data["group"], data["payload"]
        )
        self.assertIsNotNone(result1)
        result2 = bookmarks.add(
            data["owner"], data["uid"], data["group"], data["payload"]
        )
        self.assertIsNone(result2)

    def test_update_existing(self):
        from collective.bookmarks.storage import Bookmarks
        from repoze.catalog.query import Eq

        import uuid

        bookmarks = Bookmarks()
        data = {"owner": "kim stanley", "uid": uuid.uuid4(), "group": "", "payload": {}}
        bookmarks.add(data["owner"], data["uid"], data["group"], data["payload"])
        new_payload = {"foo": "bar"}
        result = bookmarks.update(
            data["owner"], data["uid"], data["group"], new_payload
        )
        self.assertIsNotNone(result)

        record = bookmarks._fetch_one(
            Eq("owner", data["owner"])
            & Eq("uid", data["uid"])
            & Eq("group", data["group"])
        )
        self.assertEqual(record.attrs["payload"], new_payload)

    def test_update_non_existing(self):
        from collective.bookmarks.storage import Bookmarks

        import uuid

        bookmarks = Bookmarks()
        data = {"owner": "kim stanley", "uid": uuid.uuid4(), "group": "", "payload": {}}
        new_payload = {"foo": "bar"}
        result = bookmarks.update(
            data["owner"], data["uid"], data["group"], new_payload
        )
        self.assertIsNone(result)

    def test_delete_existing(self):
        from collective.bookmarks.storage import Bookmarks
        from repoze.catalog.query import Eq

        import uuid

        bookmarks = Bookmarks()
        data = {"owner": "kim stanley", "uid": uuid.uuid4(), "group": "", "payload": {}}
        bookmarks.add(data["owner"], data["uid"], data["group"], data["payload"])
        result = bookmarks.delete(data["owner"], data["uid"], data["group"])
        self.assertIs(result, True)

        record = bookmarks._fetch_one(
            Eq("owner", data["owner"])
            & Eq("uid", data["uid"])
            & Eq("group", data["group"])
        )
        self.assertIsNone(record)

    def test_delete_non_existing(self):
        from collective.bookmarks.storage import Bookmarks

        import uuid

        bookmarks = Bookmarks()
        data = {"owner": "kim stanley", "uid": uuid.uuid4(), "group": "", "payload": {}}
        result = bookmarks.delete(data["owner"], data["uid"], data["group"])
        self.assertIs(result, False)

    def test_get_existing(self):
        from collective.bookmarks.storage import Bookmarks

        import uuid

        bookmarks = Bookmarks()
        data = {"owner": "kim stanley", "uid": uuid.uuid4(), "group": "", "payload": {}}
        bookmarks.add(data["owner"], data["uid"], data["group"], data["payload"])
        result = bookmarks.get(data["owner"], data["uid"], data["group"])
        result["created"] = None
        self.assertEqual(
            result,
            {
                "created": None,
                "group": "",
                "owner": "kim stanley",
                "payload": {},
                "uid": data["uid"],
            },
        )

    def test_by_owner(self):
        from collective.bookmarks.storage import Bookmarks

        import uuid

        bookmarks = Bookmarks()
        data = {"owner": "kim stanley", "uid": uuid.uuid4(), "group": "", "payload": {}}
        bookmarks.add(data["owner"], data["uid"], data["group"], data["payload"])
        bookmarks.add(data["owner"], data["uid"], "chilli", data["payload"])
        bookmarks.add("ian mcdonald", data["uid"], data["group"], data["payload"])

        result = bookmarks.by_owner(data["owner"])
        self.assertEqual(len([x for x in result]), 2)

        result = bookmarks.by_owner(data["owner"], "chilli")
        self.assertEqual(len([x for x in result]), 1)

        result = bookmarks.by_owner("non existing")
        self.assertEqual(len([x for x in result]), 0)

    def test_groups(self):
        from collective.bookmarks.storage import Bookmarks

        import uuid

        bookmarks = Bookmarks()
        data = {"owner": "kim stanley", "uid": uuid.uuid4(), "group": "", "payload": {}}
        bookmarks.add(data["owner"], data["uid"], data["group"], data["payload"])
        bookmarks.add(data["owner"], data["uid"], "chilli", data["payload"])
        bookmarks.add("ian mcdonald", data["uid"], data["group"], data["payload"])

        result = bookmarks.groups(data["owner"])
        self.assertEqual(result, ["", "chilli"])



from collective.bookmarks.api.utils import bookmark_dict_to_json_dict
from collective.bookmarks.api.utils import get_owner
from collective.bookmarks.storage import Bookmarks
from plone import api
from plone.restapi.services import Service


class BookmarksGet(Service):
    def reply(self):
        """get bookmarks by

        owner
        group (optional)
        """
        owner = get_owner(request=self.request)
        bookmarks = Bookmarks()
        bookmarks = bookmarks.by_owner(owner, group=None)
        # no bookmarks is no error.
        # if not bookmarks:
        #     raise NotFound("No such bookmark found.")
        result = []
        for bookmark in bookmarks:
            if api.content.find(UID=bookmark["uid"].hex):
                result.append(bookmark_dict_to_json_dict(bookmark))
        return result



from datetime import datetime
from plone import api
from plone.restapi.serializer.converters import json_compatible
from repoze.catalog.catalog import Catalog
from repoze.catalog.indexes.field import CatalogFieldIndex
from repoze.catalog.query import Eq
from repoze.catalog.query import Query
from souper.interfaces import ICatalogFactory
from souper.soup import get_soup
from souper.soup import NodeAttributeIndexer
from souper.soup import Record
from zope.interface import implementer

import typing
import uuid


@implementer(ICatalogFactory)
class BookmarksCatalogFactory:
    def __call__(self, context: typing.Union[Record, None] = None) -> Catalog:
        catalog = Catalog()
        catalog["uid"] = CatalogFieldIndex(NodeAttributeIndexer("uid"))
        catalog["group"] = CatalogFieldIndex(NodeAttributeIndexer("group"))
        catalog["created"] = CatalogFieldIndex(NodeAttributeIndexer("created"))
        catalog["owner"] = CatalogFieldIndex(NodeAttributeIndexer("owner"))
        catalog["queryparams"] = CatalogFieldIndex(NodeAttributeIndexer("queryparams"))
        return catalog


class Bookmarks:
    """API to manage booksmarks in the portal"""

    @property
    def _soup(self):
        """get soup storage for bookmarks"""
        soup = getattr(self, "_soup_instance", None)
        if soup is None:
            soup = get_soup("collective_bookmarks", api.portal.get())
            setattr(self, "_soup_instance", soup)
        return soup

    def _fetch_one(self, query: Query) -> typing.Union[Record, None]:
        """fetch a single record from query.

        return None if non-existing
        raises ValueError if no unique single result is found
        """
        lazy_res = self._soup.lazy(query, with_size=True)
        size = next(lazy_res)
        if size > 1:
            raise ValueError(f"Query does not point to a unique result:\n{query}")
        if size == 0:
            return None
        return next(lazy_res)()

    def _dictify(self, record: Record) -> dict:
        result = dict()
        result["owner"] = record.attrs["owner"]
        result["uid"] = record.attrs["uid"]
        result["group"] = record.attrs["group"]
        result["queryparams"] = record.attrs["queryparams"]
        result["created"] = record.attrs["created"]
        result["payload"] = record.attrs["payload"]
        return result

    def add(
        self, owner: str, uid: uuid.UUID, group: str, queryparams: str, payload: dict
    ) -> typing.Union[int, None]:
        """add new entry.

        uniqueness is given by triple of owner, uid, group and queryparams.

        returns None if such a triple already exists
        returns record_id if successful added
        """
        # check existing
        if (
            self._fetch_one(
                Eq("owner", owner)
                & Eq("uid", uid)
                & Eq("group", group)
                & Eq("queryparams", queryparams)
            )
            is not None
        ):
            return None
        record = Record()
        record.attrs["owner"] = owner
        record.attrs["uid"] = uid
        record.attrs["group"] = group
        record.attrs["queryparams"] = queryparams
        record.attrs["payload"] = payload
        record.attrs["created"] = json_compatible(datetime.utcnow())
        if self._soup.add(record):
            return self._dictify(record)

    def update(
        self, owner: str, uid: uuid.UUID, group: str, queryparams: str, payload: dict
    ) -> typing.Union[dict, None]:
        """update payload of an existing entry

        uniqueness is given by triple of owner, uid, group and queryparams.

        returns None if no such a triple already exists
        returns the Record if update was successful
        """
        record = self._fetch_one(
            Eq("owner", owner)
            & Eq("uid", uid)
            & Eq("group", group)
            & Eq("queryparams", queryparams)
        )
        if record is None:
            return None
        record.attrs["payload"] = payload
        return self._dictify(record)

    def delete(self, owner: str, uid: uuid.UUID, group: str, queryparams: str) -> bool:
        """delete existing entry

        uniqueness is given by triple of owner, uid, group and queryparams.

        returns False if no such a triple already exists
        returns True if the Record was successfully deleted
        """
        record = self._fetch_one(
            Eq("owner", owner)
            & Eq("uid", uid)
            & Eq("group", group)
            & Eq("queryparams", queryparams)
        )
        if record is None:
            return False
        del self._soup[record]
        return True

    def get(
        self, owner: str, uid: uuid.UUID, group: str, queryparams: str
    ) -> typing.Union[dict, None]:
        """get one bookmark

        uniqueness is given by triple of owner, uid, group and queryparams.

        returns None if no such a triple already exists
        returns dictified data if update was successful
        """

        record = self._fetch_one(
            Eq("owner", owner)
            & Eq("uid", uid)
            & Eq("group", group)
            & Eq("queryparams", queryparams)
        )
        if record is None:
            return None
        return self._dictify(record)

    def by_owner(
        self, owner: str, group: typing.Union[str, None] = None
    ) -> typing.Iterator[dict]:
        """get all bookmarks of an owner, optional filtered by group

        return dictified data
        """
        query = Eq("owner", owner)
        if group is not None:
            query &= Eq("group", group)
        for lazy_record in self._soup.lazy(query):
            yield self._dictify(lazy_record())

    def groups(self, owner: str) -> typing.List[str]:
        """get all groups of an owner.

        return sorted list of group names
        """
        groups = set()
        for record in self.by_owner(owner):
            groups.add(record["group"])
        return sorted(groups)



from collective.bookmarks.api.utils import get_bookmark_from_request
from collective.bookmarks.storage import Bookmarks
from plone.restapi.services import Service
from zExceptions import NotFound


class BookmarkDelete(Service):
    def reply(self):
        """delete bookmark by

        uid
        owner
        group
        queryparams (optional): serialized querystring
        """
        owner, uid, group, queryparams, payload = get_bookmark_from_request(
            self.request, loadjson=True
        )
        bookmarks = Bookmarks()
        bookmark = bookmarks.delete(owner, uid, group, queryparams)
        if bookmark:
            return self.reply_no_content()
        raise NotFound("No such bookmark found.")



"""Setup tests for this package."""
from collective.bookmarks.testing import COLLECTIVE_BOOKMARKS_INTEGRATION_TESTING
from plone import api
from plone.app.testing import setRoles
from plone.app.testing import TEST_USER_ID

import unittest


try:
    from Products.CMFPlone.utils import get_installer
except ImportError:
    get_installer = None


class TestSetup(unittest.TestCase):
    """Test that collective.bookmarks is properly installed."""

    layer = COLLECTIVE_BOOKMARKS_INTEGRATION_TESTING

    def setUp(self):
        """Custom shared utility setup for tests."""
        self.portal = self.layer["portal"]
        if get_installer:
            self.installer = get_installer(self.portal, self.layer["request"])
        else:
            self.installer = api.portal.get_tool("portal_quickinstaller")

    def test_product_installed(self):
        """Test if collective.bookmarks is installed."""
        self.assertTrue(self.installer.isProductInstalled("collective.bookmarks"))

    def test_browserlayer(self):
        """Test that ICollectiveBookmarksLayer is registered."""
        from collective.bookmarks.interfaces import ICollectiveBookmarksLayer
        from plone.browserlayer import utils

        self.assertIn(ICollectiveBookmarksLayer, utils.registered_layers())


class TestUninstall(unittest.TestCase):

    layer = COLLECTIVE_BOOKMARKS_INTEGRATION_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]
        if get_installer:
            self.installer = get_installer(self.portal, self.layer["request"])
        else:
            self.installer = api.portal.get_tool("portal_quickinstaller")
        roles_before = api.user.get_roles(TEST_USER_ID)
        setRoles(self.portal, TEST_USER_ID, ["Manager"])
        self.installer.uninstallProducts(["collective.bookmarks"])
        setRoles(self.portal, TEST_USER_ID, roles_before)

    def test_product_uninstalled(self):
        """Test if collective.bookmarks is cleanly uninstalled."""
        self.assertFalse(self.installer.isProductInstalled("collective.bookmarks"))

    def test_browserlayer_removed(self):
        """Test that ICollectiveBookmarksLayer is removed."""
        from collective.bookmarks.interfaces import ICollectiveBookmarksLayer
        from plone.browserlayer import utils

        self.assertNotIn(ICollectiveBookmarksLayer, utils.registered_layers())



from collective.bookmarks.api.utils import bookmark_dict_to_json_dict
from collective.bookmarks.api.utils import get_bookmark_from_request
from collective.bookmarks.storage import Bookmarks
from plone.protect.interfaces import IDisableCSRFProtection
from plone.restapi.services import Service
from zope.interface import alsoProvides


class BookmarkPost(Service):
    def reply(self):
        """add bookmark by

        uid
        owner
        group
        queryparams (optional): serialized querystring

        Update bookmark if bookmark with identifier set found.
        """
        owner, uid, group, queryparams, payload = get_bookmark_from_request(
            self.request, loadjson=True
        )
        alsoProvides(self.request, IDisableCSRFProtection)
        bookmarks = Bookmarks()
        bookmark = bookmarks.add(owner, uid, group, queryparams, payload)
        if not bookmark:
            # be kind
            bookmark = bookmarks.update(owner, uid, group, queryparams, payload)
        self.request.response.setStatus(201)
        return bookmark_dict_to_json_dict(bookmark)



from plone.app.contenttypes.testing import PLONE_APP_CONTENTTYPES_FIXTURE
from plone.app.testing import applyProfile
from plone.app.testing import FunctionalTesting
from plone.app.testing import IntegrationTesting
from plone.app.testing import PloneSandboxLayer

import collective.bookmarks


class CollectiveBookmarksLayer(PloneSandboxLayer):

    defaultBases = (PLONE_APP_CONTENTTYPES_FIXTURE,)

    def setUpZope(self, app, configurationContext):
        # Load any other ZCML that is required for your tests.
        # The z3c.autoinclude feature is disabled in the Plone fixture base
        # layer.
        import plone.restapi

        self.loadZCML(package=plone.restapi)
        self.loadZCML(package=collective.bookmarks)

    def setUpPloneSite(self, portal):
        applyProfile(portal, "collective.bookmarks:default")


COLLECTIVE_BOOKMARKS_FIXTURE = CollectiveBookmarksLayer()


COLLECTIVE_BOOKMARKS_INTEGRATION_TESTING = IntegrationTesting(
    bases=(COLLECTIVE_BOOKMARKS_FIXTURE,),
    name="CollectiveBookmarksLayer:IntegrationTesting",
)


COLLECTIVE_BOOKMARKS_FUNCTIONAL_TESTING = FunctionalTesting(
    bases=(COLLECTIVE_BOOKMARKS_FIXTURE,),
    name="CollectiveBookmarksLayer:FunctionalTesting",
)
